@@ -1,12 +1,12 @@
 <?php
 /**
- * Copyright 2013 Horde LLC (http://www.horde.org/)
+ * Copyright 2013-2014 Horde LLC (http://www.horde.org/)
  *
  * See the enclosed file COPYING for license information (LGPL). If you
  * did not receive this file, see http://www.horde.org/licenses/lgpl21.
  *
  * @category  Horde
- * @copyright 2013 Horde LLC
+ * @copyright 2013-2014 Horde LLC
  * @license   http://www.horde.org/licenses/lgpl21 LGPL 2.1
  * @package   Smtp
  */
@@ -29,23 +29,38 @@
  *   - RFC 5321: Simple Mail Transfer Protocol
  *   - RFC 6152/STD 71: 8bit-MIMEtransport
  *   - RFC 6409/STD 72: Message Submission for Mail
+ *   - RFC 6531: Internationalized Email
  *
  *   - XOAUTH2: https://developers.google.com/gmail/xoauth2_protocol
  * </pre>
  *
  * TODO:
  * <pre>
+ *   - RFC 1845: CHECKPOINT
+ *   - RFC 2852: DELIVERYBY
  *   - RFC 3030: BINARYMIME/CHUNKING
+ *   - RFC 3461: DSN
+ *   - RFC 3865: NO-SOLICITING
+ *   - RFC 3885: MTRK
+ *   - RFC 4141: CONPERM/CONNEG
+ *   - RFC 4405: SUBMITTER
+ *   - RFC 4468: BURL
+ *   - RFC 4865: FUTURERELEASE
+ *   - RFC 6710: MT-PRIORITY
+ *   - RFC 7293: RRVS
  * </pre>
  *
  * @author    Michael Slusarz <slusarz@horde.org>
  * @category  Horde
- * @copyright 2013 Horde LLC
+ * @copyright 2013-2014 Horde LLC
  * @license   http://www.horde.org/licenses/lgpl21 LGPL 2.1
  * @package   Smtp
  *
  * @property-read boolean $data_8bit  Does server support sending 8-bit MIME
  *                                    data?
+ * @property-read boolean $data_intl  Does server support sending
+ *                                    internationalized (UTF-8) header data?
+ *                                    (@since 1.6.0)
  * @property-read integer $size  The maximum message size supported (in
  *                               bytes) or null if this cannot be determined.
  */
@@ -66,7 +81,14 @@ class Horde_Smtp implements Serializable
     protected $_debug;
 
     /**
-     * The list of extensions.
+     * The hello command to use for extended SMTP support.
+     *
+     * @var string
+     */
+    protected $_ehlo = 'EHLO';
+
+    /**
+     * The list of ESMTP extensions.
      * If this value is null, we have not connected to server yet.
      *
      * @var array
@@ -80,6 +102,13 @@ class Horde_Smtp implements Serializable
      */
     protected $_params = array();
 
+    /**
+     * List of required ESMTP extensions.
+     *
+     * @var array
+     */
+    protected $_requiredExts = array();
+
     /**
      * Constructor.
      *
@@ -102,7 +131,7 @@ class Horde_Smtp implements Serializable
      *  </li>
      *  <li>
      *   port: (string) The SMTP port.
-     *         DEFAULT: 25
+     *         DEFAULT: 587 (See RFC 6409/STD 72)
      *  </li>
      *  <li>
      *   secure: (string) Use SSL or TLS to connect.
@@ -140,7 +169,7 @@ class Horde_Smtp implements Serializable
         // Default values.
         $params = array_merge(array(
             'host' => 'localhost',
-            'port' => 25,
+            'port' => 587,
             'secure' => true,
             'timeout' => 30
         ), array_filter($params));
@@ -219,6 +248,10 @@ class Horde_Smtp implements Serializable
             // RFC 6152
             return $this->queryExtension('8BITMIME');
 
+        case 'data_intl':
+            // RFC 6531
+            return $this->queryExtension('SMTPUTF8');
+
         case 'size':
             // RFC 1870
             return $this->queryExtension('SIZE') ?: null;
@@ -238,7 +271,8 @@ class Horde_Smtp implements Serializable
         switch ($key) {
         case 'password':
         case 'xoauth2_token':
-            if ($this->_params[$key] instanceof Horde_Smtp_Password) {
+            if (isset($this->_params[$key]) &&
+                ($this->_params[$key] instanceof Horde_Smtp_Password)) {
                 return $this->_params[$key]->getPassword();
             }
 
@@ -347,52 +381,24 @@ class Horde_Smtp implements Serializable
                 );
             } catch (Horde\Socket\Client\Exception $e) {
                 $e2 = new Horde_Smtp_Exception(
-                    Horde_Smtp_Translation::t("Error connecting to SMTP server."),
+                    Horde_Smtp_Translation::r("Error connecting to SMTP server."),
                     Horde_Smtp_Exception::SERVER_CONNECT
                 );
                 $e2->details = $e->details;
                 throw $e2;
             }
 
+            $this->_debug->info(sprintf(
+                'Connection to: smtp://%s:%s',
+                $this->getParam('host'),
+                $this->getParam('port')
+            ));
+
             // Get initial line (RFC 5321 [3.1]).
             $this->_getResponse(220, 'logout');
         }
 
-        $ehlo = $host = gethostname();
-        if ($host === false) {
-            $ehlo = $_SERVER['SERVER_ADDR'];
-            $host = 'localhost';
-        }
-
-        $this->_connection->write('EHLO ' . $ehlo);
-        try {
-            $resp = $this->_getResponse(250);
-
-            foreach ($resp as $val) {
-                $tmp = explode(' ', $val, 2);
-                $this->_extensions[$tmp[0]] = empty($tmp[1])
-                    ? true
-                    : $tmp[1];
-            }
-        } catch (Horde_Smtp_Exception $e) {
-            switch ($e->getSmtpCode()) {
-            case 502:
-                // Old server - doesn't support EHLO
-                $this->_connection->write('HELO ' . $host);
-                try {
-                    $this->_getResponse(250);
-                } catch (Horde_Smtp_Exception $e2) {
-                    $this->logout();
-                    throw $e;
-                }
-                $this->_extensions = array();
-                break;
-
-            default:
-                $this->logout();
-                throw $e;
-            }
-        }
+        $this->_hello();
 
         if ($this->_startTls()) {
             $this->_extensions = null;
@@ -400,6 +406,19 @@ class Horde_Smtp implements Serializable
             return;
         }
 
+        /* Check for required ESMTP extensions. */
+        foreach ($this->_requiredExts as $val) {
+            if (!$this->queryExtension($val)) {
+                throw new Horde_Smtp_Exception(
+                    sprintf(
+                        Horde_Smtp_Translation::r("Server does not support a necessary server extension: %s."),
+                        $val
+                    ),
+                    Horde_Smtp_Exception::LOGIN_MISSINGEXTENSION
+                );
+            }
+        }
+
         /* If we reached this point and don't have a secure connection, then
          * a secure connections is not available. */
         if (!$this->isSecureConnection() &&
@@ -431,7 +450,7 @@ class Horde_Smtp implements Serializable
 
         $this->logout();
         throw new Horde_Smtp_Exception(
-            Horde_Smtp_Translation::t("Server denied authentication."),
+            Horde_Smtp_Translation::r("Server denied authentication."),
             Horde_Smtp_Exception::LOGIN_AUTHENTICATIONFAILED
         );
     }
@@ -469,8 +488,19 @@ class Horde_Smtp implements Serializable
      *   - 8bit: (boolean) If true, $data is a MIME message with arbitrary
      *           octet content (i.e. 8-bit encoding).
      *           DEFAULT: false
+     *   - intl: (boolean) If true, $data contains internationalized header
+     *           content (UTF-8). (@since 1.6.0)
+     *           DEFAULT: false
      * </pre>
      *
+     * @return array  If no receipients were successful, a
+     *                Horde_Smtp_Exception will be thrown. If at least one
+     *                recipient was successful, an array with the following
+     *                format is returned: (@since 1.5.0)
+     *   - KEYS: Recipient addresses ($to addresses).
+     *   - VALUES: Boolean true (if message was accepted for this recpieint)
+     *             or a Horde_Smtp_Exception (if messages was not accepted).
+     *
      * @throws Horde_Smtp_Exception
      */
     public function send($from, $to, $data, array $opts = array())
@@ -481,7 +511,28 @@ class Horde_Smtp implements Serializable
             $from = new Horde_Mail_Rfc822_Address($from);
         }
 
-        $mailcmd = 'MAIL FROM:<' . $from->bare_address_idn . '>';
+        /* RFC 6531 */
+        if (!empty($opts['intl'])) {
+            if (!$this->data_intl) {
+                throw new InvalidArgumentException(
+                    'Server does not support sending internationalized header data.'
+                );
+            }
+
+            /* RFC 6531[1.2] requires 8BITMIME to be available. */
+            $opts['8bit'] = true;
+        }
+
+        /* RFC 6152[3] */
+        if (!empty($opts['8bit']) && !$this->data_8bit) {
+            throw new InvalidArgumentException(
+                'Server does not support sending 8-bit data.'
+            );
+        }
+
+        $mailcmd = 'MAIL FROM:<' .
+            (empty($opts['intl']) ? $from->bare_address_idn : $from->bare_address) .
+            '>';
 
         // RFC 1870[6]
         if ($this->queryExtension('SIZE')) {
@@ -495,9 +546,18 @@ class Horde_Smtp implements Serializable
             $mailcmd .= ' SIZE=' . intval($size);
         }
 
+        // RFC 6531[3.4]
+        if (!empty($opts['intl'])) {
+            $mailcmd .= ' SMTPUTF8';
+        }
+
         // RFC 6152[3]
         if (!empty($opts['8bit'])) {
             $mailcmd .= ' BODY=8BITMIME';
+        } elseif ($this->_debug->active && $this->data_8bit) {
+            /* Only output extended 7bit command if debug is active (it is
+             * default and does not need to be explicitly declared). */
+            $mailcmd .= ' BODY=7BIT';
         }
 
         $cmds = array($mailcmd);
@@ -506,7 +566,10 @@ class Horde_Smtp implements Serializable
             $to = new Horde_Mail_Rfc822_List($to);
         }
 
-        foreach ($to->bare_addresses_idn as $val) {
+        $recipients = empty($opts['intl'])
+            ? $to->bare_addresses_idn
+            : $to->bare_addresses;
+        foreach ($recipients as $val) {
             $cmds[] = 'RCPT TO:<' . $val . '>';
         }
 
@@ -573,7 +636,8 @@ class Horde_Smtp implements Serializable
         stream_filter_remove($res);
 
         $this->_connection->write('.');
-        $this->_getResponse(250, 'reset');
+
+        return $this->_processData($recipients);
     }
 
     /**
@@ -612,6 +676,50 @@ class Horde_Smtp implements Serializable
 
     /* Internal methods. */
 
+    /**
+     * Send "Hello" command to the server.
+     *
+     * @throws Horde_Smtp_Exception
+     */
+    protected function _hello()
+    {
+        $ehlo = $host = gethostname();
+        if ($host === false) {
+            $ehlo = $_SERVER['SERVER_ADDR'];
+            $host = 'localhost';
+        }
+
+        $this->_connection->write($this->_ehlo . ' ' . $ehlo);
+        try {
+            $resp = $this->_getResponse(250);
+
+            foreach ($resp as $val) {
+                $tmp = explode(' ', $val, 2);
+                $this->_extensions[$tmp[0]] = empty($tmp[1])
+                    ? true
+                    : $tmp[1];
+            }
+        } catch (Horde_Smtp_Exception $e) {
+            switch ($e->getSmtpCode()) {
+            case 502:
+                // Old server - doesn't support EHLO
+                $this->_connection->write('HELO ' . $host);
+                try {
+                    $this->_getResponse(250);
+                } catch (Horde_Smtp_Exception $e2) {
+                    $this->logout();
+                    throw $e;
+                }
+                $this->_extensions = array();
+                break;
+
+            default:
+                $this->logout();
+                throw $e;
+            }
+        }
+    }
+
     /**
      * Starts the TLS connection to the server, if necessary.  See RFC 3207.
      *
@@ -634,7 +742,7 @@ class Horde_Smtp implements Serializable
             }
 
             throw new Horde_Smtp_Exception(
-                Horde_Smtp_Translation::t("Server does not support TLS connections."),
+                Horde_Smtp_Translation::r("Server does not support TLS connections."),
                 Horde_Smtp_Exception::LOGIN_TLSFAILURE
             );
         }
@@ -828,4 +936,20 @@ class Horde_Smtp implements Serializable
         throw $e;
     }
 
+    /**
+     * Process the return from the DATA command.
+     *
+     * @see _send()
+     *
+     * @param array $recipients  The list of message recipients.
+     *
+     * @return array  See _send().
+     * @throws Horde_Smtp_Exception
+     */
+    protected function _processData($recipients)
+    {
+        $this->_getResponse(250, 'reset');
+        return array_fill_keys($recipients, true);
+    }
+
 }
